# -*- coding: utf-8 -*-
import cmath
from typing import Dict, List, Tuple

import numpy as np
import ray
from scipy.special import jn_zeros, jnp_zeros, jv, jvp, kv, kvp

from pymwm.utils.cylinder_utils import ABY_cython, coefs_cython, uvABY_cython
from pymwm.waveguide import Waveguide

from .samples import Samples, SamplesForRay, SamplesLowLoss, SamplesLowLossForRay


class Cylinder(Waveguide):
    """A class defining a cylindrical waveguide."""

    def __init__(self, params):
        """Init Cylinder class.

        Args:
            params: A dict whose keys and values are as follows:
                'core': A dict of the setting parameters of the core:
                    'shape': A string indicating the shape of the core.
                    'size': A float indicating the radius of the circular cross
                        section [um].
                    'fill': A dict of the parameters of the core Material.
                'clad': A dict of the parameters of the clad Material.
                'bounds': A dict indicating the bounds of database.interpolation
                    and its keys and values are as follows:
                    'wl_max': A float indicating the maximum wavelength [um]
                    'wl_min': A float indicating the minimum wavelength [um]
                    'wl_imag': A float indicating the maximum value of
                        abs(c / f_imag) [um] where f_imag is the imaginary part
                        of the frequency.
                'modes': A dict of the settings for calculating modes:
                    'wl_max': A float indicating the maximum wavelength [um]
                        (default: 5.0)
                    'wl_min': A float indicating the minimum wavelength [um]
                        (default: 0.4)
                    'wl_imag': A float indicating the maximum value of
                        abs(c / f_imag) [um] where f_imag is the imaginary part
                        of the frequency. (default: 5.0)
                    'dw': A float indicating frequency interval
                        [rad c / 1um]=[2.99792458e14 rad / s]
                        (default: 1 / 64).
                    'num_n': An integer indicating the number of orders of
                        modes.
                    'num_m': An integer indicating the number of modes in each
                        order and polarization.
                    'ls': A list of characters chosen from "h" (horizontal
                        polarization) and "v" (vertical polarization).
        """
        super().__init__(params)
        self.u_pec, self.jnu_pec, self.jnpu_pec = self.u_jnu_jnpu_pec(
            self.num_n, self.num_m
        )

    def get_alphas(self, alpha_list: List[Tuple[str, int, int]]) -> Dict:
        alphas = {"h": [], "v": []}
        for alpha in [("E", 0, m) for m in range(1, self.num_m + 1)]:
            if alpha in alpha_list:
                alphas["v"].append(alpha)
        for alpha in [
            ("E", n, m) for n in range(1, self.num_n) for m in range(1, self.num_m + 1)
        ]:
            if alpha in alpha_list:
                alphas["h"].append(alpha)
                alphas["v"].append(alpha)
        for alpha in [("M", 0, m) for m in range(1, self.num_m + 1)]:
            if alpha in alpha_list:
                alphas["h"].append(alpha)
        for alpha in [
            ("M", n, m) for n in range(1, self.num_n) for m in range(1, self.num_m + 1)
        ]:
            if alpha in alpha_list:
                alphas["h"].append(alpha)
                alphas["v"].append(alpha)
        return alphas

    def betas_convs_samples(
        self, params: Dict
    ) -> Tuple[np.ndarray, np.ndarray, Samples]:
        im_factor = self.clad.im_factor
        self.clad.im_factor = 1.0
        p_modes = params["modes"].copy()
        num_n_0 = p_modes["num_n"]
        num_m_0 = p_modes["num_m"]
        betas = convs = None
        success = False
        for num_n, num_m in [
            (n, m) for n in range(num_n_0, 17) for m in range(num_m_0, 5)
        ]:
            p_modes["num_n"] = num_n
            p_modes["num_m"] = num_m
            smp = Samples(self.r, self.fill_params, self.clad_params, p_modes)
            try:
                betas, convs = smp.database.load()
                success = True
                break
            except IndexError:
                continue
        if not success:
            p_modes["num_n"] = num_n_0
            p_modes["num_m"] = num_m_0
            smp = Samples(self.r, self.fill_params, self.clad_params, p_modes)
            ray.shutdown()
            try:
                ray.init()
                p_modes_id = ray.put(p_modes)
                pool = ray.util.ActorPool(
                    SamplesForRay.remote(
                        self.r, self.fill_params, self.clad_params, p_modes_id
                    )
                    for _ in range(num_n_0)
                )
                xs_success_list = list(
                    pool.map(lambda a, arg: a.task.remote(arg), range(num_n_0))
                )
            finally:
                ray.shutdown()
            betas, convs = smp.betas_convs(xs_success_list)
            smp.database.save(betas, convs)
        if im_factor != 1.0:
            self.clad.im_factor = im_factor
            smp = SamplesLowLoss(self.r, self.fill_params, self.clad_params, p_modes)
            try:
                betas, convs = smp.database.load()
            except IndexError:
                self.clad.im_factor = im_factor
                num_n = p_modes["num_n"]
                num_m = p_modes["num_m"]
                args = []
                for iwr in range(len(smp.ws)):
                    for iwi in range(len(smp.wis)):
                        xis_list = []
                        for n in range(num_n):
                            xis = []
                            for i in range(num_m + 1):
                                xis.append(betas[("M", n, i + 1)][iwr, iwi] ** 2)
                            for i in range(num_m):
                                xis.append(betas[("E", n, i + 1)][iwr, iwi] ** 2)
                            xis_list.append(xis)
                        args.append((iwr, iwi, xis_list))
                try:
                    ray.init()
                    p_modes_id = ray.put(p_modes)
                    pool = ray.util.ActorPool(
                        SamplesLowLossForRay.remote(
                            self.r, self.fill_params, self.clad_params, p_modes_id
                        )
                        for _ in range(16)
                    )
                    xs_success_list = list(
                        pool.map(lambda a, arg: a.task.remote(arg), args)
                    )
                finally:
                    ray.shutdown()
                betas, convs = smp.betas_convs(xs_success_list)
                smp.database.save(betas, convs)
        return betas, convs, smp

    def beta(self, w: complex, alpha: Tuple[str, int, int]) -> complex:
        """Return phase constant

        Args:
            w: A complex indicating the angular frequency
            alpha: (pol, n, m)
                pol: 'M' (TM-like mode) or 'E' (TE-like mode)
                n: The order of the mode
                m: The sub order of the mode.
        Returns:
            h: The phase constant.
        """
        wr = w.real
        wi = w.imag
        hr = self.beta_funcs[(alpha, "real")](wr, wi)[0, 0]
        hi = self.beta_funcs[(alpha, "imag")](wr, wi)[0, 0]
        # if hr < 0:
        #     hr = 1e-16
        # if hi < 0:
        #     hi = 1e-16
        return hr + 1j * hi

    def beta_pec(self, w, alpha):
        """Return phase constant of PEC waveguide

        Args:
            w: A complex indicating the angular frequency
            alpha: A tuple (pol, n, m) where pol is 'M' for TM mode or
                'E' for TE mode, n is the order of the mode, and m is the
                number of modes in the order and the polarization.
        Returns:
            h: A complex indicating the phase constant.
        """
        w_comp = w.real + 1j * w.imag
        pol, n, m = alpha
        if pol == "E":
            chi = jnp_zeros(n, m)[-1]
        elif pol == "M":
            chi = jn_zeros(n, m)[-1]
        else:
            raise ValueError("pol must be 'E' or 'M")
        val = cmath.sqrt(self.fill(w_comp) * w_comp ** 2 - chi ** 2 / self.r ** 2)
        if abs(val.real) > abs(val.imag):
            if val.real < 0:
                val *= -1
        else:
            if val.imag < 0:
                val *= -1
        return val

    def coef(self, h, w, alpha):
        """Return the coefficients of TE- and TM- components which compose
        the hybrid mode.

        Args:
            h: A complex indicating the phase constant.
            w: A complex indicating the angular frequency
            alpha: A tuple (pol, n, m) where pol is 'M' for TM-like mode or
                'E' for TE-like mode, n is the order of the mode, and m is
                the number of modes in the order and the polarization.
        Returns:
            a: A complex indicating the coefficient of TE-component
            b: A complex indicating the coefficient of TM-component
        """
        e1 = self.fill(w)
        e2 = self.clad(w)
        pol, n, m = alpha
        w = w.real + 1j * w.imag
        h = h.real + 1j * h.imag
        if e2.real < -1e6:
            if pol == "E":
                norm = self.norm(w, h, alpha, 1.0 + 0.0j, 0.0j)
                ai, bi = 1.0 / norm, 0.0
            else:
                norm = self.norm(w, h, alpha, 0.0j, 1.0 + 0.0j)
                ai, bi = 0.0, 1.0 / norm
        else:
            u = self.samples.u(h ** 2, w, e1)
            v = self.samples.v(h ** 2, w, e2)
            knv = kv(n, v)
            knpv = kvp(n, v)
            jnu = jv(n, u)
            jnpu = jvp(n, u)
            ci = -n * (u ** 2 + v ** 2) * jnu * knv / (u * v)
            if pol == "E":
                ci *= (h / w) ** 2
                ci /= e1 * jnpu * v * knv + e2 * knpv * u * jnu
                norm = self.norm(w, h, alpha, 1.0 + 0.0j, ci)
                ai = 1.0 / norm
                bi = ci / norm
            else:
                ci /= jnpu * v * knv + knpv * u * jnu
                norm = self.norm(w, h, alpha, ci, 1.0 + 0.0j)
                bi = 1.0 / norm
                ai = ci / norm
        return ai, bi

    def norm(self, w, h, alpha, a, b):
        pol, n, m = alpha
        en = 1 if n == 0 else 2
        if self.clad(w).real < -1e6:
            radius = self.r
            if pol == "E":
                u = jnp_zeros(n, m)[-1]
                jnu = jv(n, u)
                jnpu = 0.0
            else:
                u = jn_zeros(n, m)[-1]
                jnu = 0.0
                jnpu = jvp(n, u)
            return cmath.sqrt(
                a ** 2 * np.pi * radius ** 2 / en * (1 - n ** 2 / u ** 2) * jnu ** 2
                + b ** 2 * np.pi * radius ** 2 / en * jnpu ** 2
            )
        # ac = a.conjugate()
        # bc = b.conjugate()
        u = self.samples.u(h ** 2, w, self.fill(w))
        jnu = jv(n, u)
        jnpu = jvp(n, u)
        v = self.samples.v(h ** 2, w, self.clad(w))
        knv = kv(n, v)
        knpv = kvp(n, v)
        # uc = u.conjugate()
        # jnuc = jnu.conjugate()
        # jnpuc = jnpu.conjugate()
        # vc = v.conjugate()
        # knvc = knv.conjugate()
        # knpvc = knpv.conjugate()
        val_u = 2 * np.pi * self.r ** 2 / en
        val_v = val_u * ((u * jnu) / (v * knv)) ** 2
        # val_v = val_u * (uc * u * jnuc * jnu) / (vc * v * knvc * knv)
        upart_diag = self.upart_diag(n, u, jnu, jnpu, u, jnu, jnpu)
        vpart_diag = self.vpart_diag(n, v, knv, knpv, v, knv, knpv)
        upart_off = self.upart_off(n, u, jnu, u, jnu)
        vpart_off = self.vpart_off(n, v, knv, v, knv)
        return cmath.sqrt(
            val_u
            * (
                a * (a * upart_diag + b * upart_off)
                + b * (b * upart_diag + a * upart_off)
            )
            - val_v
            * (
                a * (a * vpart_diag + b * vpart_off)
                + b * (b * vpart_diag + a * vpart_off)
            )
        )
        # upart_diag = self.upart_diag(n, uc, jnuc, jnpuc, u, jnu, jnpu)
        # vpart_diag = self.vpart_diag(n, vc, knvc, knpvc, v, knv, knpv)
        # upart_off = self.upart_off(n, uc, jnuc, u, jnu)
        # vpart_off = self.vpart_off(n, vc, knvc, v, knv)
        # return cmath.sqrt(np.real(
        #     val_u * (
        #         a * (ac * upart_diag + bc * upart_off) +
        #         b * (bc * upart_diag + ac * upart_off)) -
        #     val_v * (
        #         a * (ac * vpart_diag + bc * vpart_off) +
        #         b * (bc * vpart_diag + ac * vpart_off))))

    @staticmethod
    def upart_diag(n, uc, jnuc, jnpuc, u, jnu, jnpu):
        if abs(uc - u) < 1e-10:
            u0 = (u + uc) / 2
            jnu0 = jv(n, u0)
            jnpu0 = jvp(n, u0)
            return (
                jnu0 * jnpu0 / u0
                + (jnpu0 ** 2 + (1 - n ** 2 / u0 ** 2) * jnu0 ** 2) / 2
            )
        if abs(uc + u) < 1e-10:
            u0 = (u - uc) / 2
            jnu0 = jv(n, u0)
            jnpu0 = jvp(n, u0)
            return (-1) ** (n - 1) * (
                jnu0 * jnpu0 / u0
                + (jnpu0 ** 2 + (1 - n ** 2 / u0 ** 2) * jnu0 ** 2) / 2
            )
        return (uc * jnuc * jnpu - u * jnu * jnpuc) / (uc ** 2 - u ** 2)

    @staticmethod
    def upart_off(n, uc, jnuc, u, jnu):
        return n * (jnuc * jnu) / (uc * u)

    @staticmethod
    def vpart_diag(n, vc, knvc, knpvc, v, knv, knpv):
        if abs(vc - v) < 1e-10:
            v0 = (v + vc) / 2
            knv0 = kv(n, v0)
            knpv0 = kvp(n, v0)
            return (
                knv0 * knpv0 / v0
                + (knpv0 ** 2 - (1 + n ** 2 / v0 ** 2) * knv0 ** 2) / 2
            )
        if abs(vc + v) < 1e-10:
            v0 = (v - vc) / 2
            knv0 = kv(n, v0)
            knpv0 = kvp(n, v0)
            return (-1) ** (n - 1) * (
                knv0 * knpv0 / v0
                + (knpv0 ** 2 - (1 + n ** 2 / v0 ** 2) * knv0 ** 2) / 2
            )
        return (vc * knvc * knpv - v * knv * knpvc) / (vc ** 2 - v ** 2)

    @staticmethod
    def vpart_off(n, vc, knvc, v, knv):
        return n * (knvc * knv) / (vc * v)

    def Y(self, w, h, alpha, a, b):
        """Return the effective admittance of the waveguide mode

        Args:
            w: A complex indicating the angular frequency
            h: A complex indicating the phase constant.
            alpha: A tuple (pol, n, m) where pol is 'M' for TM-like mode or
                'E' for TE-like mode, n is the order of the mode, and m is
                the number of modes in the order and the polarization.
            a: A complex indicating the coefficient of TE-component
            b: A complex indicating the coefficient of TM-component
        Returns:
            y: A complex indicating the effective admittance
        """
        pol, n, m = alpha
        e1 = self.fill(w)
        e2 = self.clad(w)
        en = 1 if n == 0 else 2
        if e2.real < -1e6:
            if pol == "E":
                val = h / w
            else:
                val = e1 * w / h
        else:
            u = self.samples.u(h ** 2, w, e1)
            jnu = jv(n, u)
            jnpu = jvp(n, u)
            v = self.samples.v(h ** 2, w, e2)
            knv = kv(n, v)
            knpv = kvp(n, v)
            val_u = 2 * np.pi * self.r ** 2 / en
            val_v = val_u * ((u * jnu) / (v * knv)) ** 2
            upart_diag = self.upart_diag(n, u, jnu, jnpu, u, jnu, jnpu)
            vpart_diag = self.vpart_diag(n, v, knv, knpv, v, knv, knpv)
            upart_off = self.upart_off(n, u, jnu, u, jnu)
            vpart_off = self.vpart_off(n, v, knv, v, knv)
            val = val_u * (
                h / w * a * (a * upart_diag + b * upart_off)
                + e1 * w / h * b * (b * upart_diag + a * upart_off)
            ) - val_v * (
                h / w * a * (a * vpart_diag + b * vpart_off)
                + e2 * w / h * b * (b * vpart_diag + a * vpart_off)
            )
        return val

    def Yab(self, w, h1, s1, l1, n1, m1, a1, b1, h2, s2, l2, n2, m2, a2, b2):
        """Return the admittance matrix element of the waveguide modes using
        orthogonality.

        Args:
            w: A complex indicating the angular frequency
            h1: A complex indicating the phase constant.
            s1: 0 for TE-like mode or 1 for TM-like mode
            l1: 0 for h mode or 1 for v mode
            n1: the order of the mode
            m1: the number of modes in the order and the polarization
            a1: A complex indicating the coefficient of TE-component
            b1: A complex indicating the coefficient of TM-component
            h2: A complex indicating the phase constant.
            s2: 0 for TE-like mode or 1 for TM-like mode
            l2: 0 for h mode or 1 for v mode
            n2: the order of the mode
            m2: the number of modes in the order and the polarization
            a2: A complex indicating the coefficient of TE-component
            b2: A complex indicating the coefficient of TM-component
        Returns:
            y: A complex indicating the effective admittance
        """
        if n1 != n2 or l1 != l2:
            return 0.0
        n = n1
        e1 = self.fill(w)
        e2 = self.clad(w)
        en = 1 if n == 0 else 2
        if e2.real < -1e6:
            if s1 != s2 or m1 != m2:
                return 0.0
            if s1 == 0:
                val = h2 / w
            else:
                val = e1 * w / h2
        else:
            ac = a1
            bc = b1
            a, b = a2, b2
            uc = self.samples.u(h1 ** 2, w, e1)
            vc = self.samples.v(h1 ** 2, w, e2)
            u = self.samples.u(h2 ** 2, w, e1)
            v = self.samples.v(h2 ** 2, w, e2)
            jnuc = jv(n, uc)
            jnpuc = jvp(n, uc)
            knvc = kv(n, vc)
            knpvc = kvp(n, vc)
            jnu = jv(n, u)
            jnpu = jvp(n, u)
            knv = kv(n, v)
            knpv = kvp(n, v)
            val_u = 2 * np.pi * self.r ** 2 / en
            val_v = val_u * (uc * u * jnuc * jnu) / (vc * v * knvc * knv)
            upart_diag = self.upart_diag(n, uc, jnuc, jnpuc, u, jnu, jnpu)
            vpart_diag = self.vpart_diag(n, vc, knvc, knpvc, v, knv, knpv)
            upart_off = self.upart_off(n, uc, jnuc, u, jnu)
            vpart_off = self.vpart_off(n, vc, knvc, v, knv)
            val = val_u * (
                h2 / w * a * (ac * upart_diag + bc * upart_off)
                + e1 * w / h2 * b * (bc * upart_diag + ac * upart_off)
            ) - val_v * (
                h2 / w * a * (ac * vpart_diag + bc * vpart_off)
                + e2 * w / h2 * b * (bc * vpart_diag + ac * vpart_off)
            )
        return val

    @staticmethod
    def y_te(w, h):
        return h / w

    def y_tm_inner(self, w, h):
        e = self.fill(w)
        return e * w / h

    def y_tm_outer(self, w, h):
        e = self.clad(w)
        return e * w / h

    def fields(self, x, y, w, dir, alpha, h, coef):
        """Return the electromagnetic field vectors for the specified mode and
        point

        Args:
            x: A float indicating the x coordinate [um]
            y: A float indicating the y coordinate [um]
            w: A complex indicating the angular frequency
            dir: "h" (horizontal polarization) or "v" (vertical polarization)
            alpha: A tuple (pol, n, m) where pol is 'M' for TM-like mode or
                'E' for TE-like mode, n is the order of the mode, and m is
                the number of modes in the order and the polarization.
            h: A complex indicating the phase constant.
            coef: The coefficients of TE- and TM- components
        Returns:
            f_vec: An array of complexes [ex, ey, ez, hx, hy, hz].
        """
        pol, n, m = alpha
        a, b = coef
        r = np.hypot(x, y)
        p = np.arctan2(y, x)
        u = self.samples.u(h ** 2, w, self.fill(w))
        v = self.samples.v(h ** 2, w, self.clad(w))
        ur = u * r / self.r
        vr = v * r / self.r
        if dir == "h":
            fr = np.cos(n * p)
            fp = -np.sin(n * p)
        else:
            fr = np.sin(n * p)
            fp = np.cos(n * p)
        y_te = self.y_te(w, h)
        if r <= self.r:
            y_tm = self.y_tm_inner(w, h)
            er_te = (jv(n - 1, ur) + jv(n + 1, ur)) / 2 * fr
            er_tm = jvp(n, ur) * fr
            er = a * er_te + b * er_tm
            ep_te = jvp(n, ur) * fp
            ep_tm = (jv(n - 1, ur) + jv(n + 1, ur)) / 2 * fp
            ep = a * ep_te + b * ep_tm
            ez = u / (1j * h * self.r) * b * jv(n, ur) * fr
            hr = -y_te * a * ep_te - y_tm * b * ep_tm
            hp = y_te * a * er_te + y_tm * b * er_tm
            hz = -u / (1j * h * self.r) * y_te * a * jv(n, ur) * fp
        else:
            y_tm = self.y_tm_outer(w, h)
            val = -u * jv(n, u) / (v * kv(n, v))
            er_te = -(kv(n - 1, vr) - kv(n + 1, vr)) / 2 * fr * val
            er_tm = kvp(n, vr) * fr * val
            er = a * er_te + b * er_tm
            ep_te = kvp(n, vr) * fp * val
            ep_tm = -(kv(n - 1, vr) - kv(n + 1, vr)) / 2 * fp * val
            ep = a * ep_te + b * ep_tm
            ez = -v / (1j * h * self.r) * b * kv(n, vr) * fr * val
            hr = -y_te * a * ep_te - y_tm * b * ep_tm
            hp = y_te * a * er_te + y_tm * b * er_tm
            hz = v / (1j * h * self.r) * y_te * a * kv(n, vr) * fp * val
        ex = er * np.cos(p) - ep * np.sin(p)
        ey = er * np.sin(p) + ep * np.cos(p)
        hx = hr * np.cos(p) - hp * np.sin(p)
        hy = hr * np.sin(p) + hp * np.cos(p)
        return np.array([ex, ey, ez, hx, hy, hz])

    def e_field(self, x, y, w, dir, alpha, h, coef):
        """Return the electric field vector for the specified mode and
        point

        Args:
            x: A float indicating the x coordinate [um]
            y: A float indicating the y coordinate [um]
            w: A complex indicating the angular frequency
            dir: "h" (horizontal polarization) or "v" (vertical polarization)
            alpha: A tuple (pol, n, m) where pol is 'M' for TM-like mode or
                'E' for TE-like mode, n is the order of the mode, and m is
                the number of modes in the order and the polarization.
            h: A complex indicating the phase constant.
            coef: The coefficients of TE- and TM- components
        Returns:
            e_vec: An array of complexes [ex, ey, ez].
        """
        pol, n, m = alpha
        a, b = coef
        r = np.hypot(x, y)
        p = np.arctan2(y, x)
        u = self.samples.u(h ** 2, w, self.fill(w))
        v = self.samples.v(h ** 2, w, self.clad(w))
        ur = u * r / self.r
        vr = v * r / self.r
        if dir == "h":
            fr = np.cos(n * p)
            fp = -np.sin(n * p)
        else:
            fr = np.sin(n * p)
            fp = np.cos(n * p)
        if r <= self.r:
            er_te = (jv(n - 1, ur) + jv(n + 1, ur)) / 2 * fr
            er_tm = jvp(n, ur) * fr
            er = a * er_te + b * er_tm
            ep_te = jvp(n, ur) * fp
            ep_tm = (jv(n - 1, ur) + jv(n + 1, ur)) / 2 * fp
            ep = a * ep_te + b * ep_tm
            ez = u / (1j * h * self.r) * b * jv(n, ur) * fr
        else:
            val = -u * jv(n, u) / (v * kv(n, v))
            er_te = -(kv(n - 1, vr) - kv(n + 1, vr)) / 2 * fr * val
            er_tm = kvp(n, vr) * fr * val
            er = a * er_te + b * er_tm
            ep_te = kvp(n, vr) * fp * val
            ep_tm = -(kv(n - 1, vr) - kv(n + 1, vr)) / 2 * fp * val
            ep = a * ep_te + b * ep_tm
            ez = -v / (1j * h * self.r) * b * kv(n, vr) * fr * val
        ex = er * np.cos(p) - ep * np.sin(p)
        ey = er * np.sin(p) + ep * np.cos(p)
        return np.array([ex, ey, ez])

    def h_field(self, x, y, w, dir, alpha, h, coef):
        """Return the magnetic field vectors for the specified mode and
        point

        Args:
            x: A float indicating the x coordinate [um]
            y: A float indicating the y coordinate [um]
            w: A complex indicating the angular frequency
            dir: "h" (horizontal polarization) or "v" (vertical polarization)
            alpha: A tuple (pol, n, m) where pol is 'M' for TM-like mode or
                'E' for TE-like mode, n is the order of the mode, and m is
                the number of modes in the order and the polarization.
            h: A complex indicating the phase constant.
            coef: The coefficients of TE- and TM- components
        Returns:
            h_vec: An array of complexes [hx, hy, hz].
        """
        pol, n, m = alpha
        a, b = coef
        r = np.hypot(x, y)
        p = np.arctan2(y, x)
        u = self.samples.u(h ** 2, w, self.fill(w))
        v = self.samples.v(h ** 2, w, self.clad(w))
        ur = u * r / self.r
        vr = v * r / self.r
        if dir == "h":
            fr = np.cos(n * p)
            fp = -np.sin(n * p)
        else:
            fr = np.sin(n * p)
            fp = np.cos(n * p)
        y_te = self.y_te(w, h)
        if r <= self.r:
            y_tm = self.y_tm_inner(w, h)
            er_te = (jv(n - 1, ur) + jv(n + 1, ur)) / 2 * fr
            er_tm = jvp(n, ur) * fr
            ep_te = jvp(n, ur) * fp
            ep_tm = (jv(n - 1, ur) + jv(n + 1, ur)) / 2 * fp
            hr = -y_te * a * ep_te - y_tm * b * ep_tm
            hp = y_te * a * er_te + y_tm * b * er_tm
            hz = -u / (1j * h * self.r) * y_te * a * jv(n, ur) * fp
        else:
            y_tm = self.y_tm_outer(w, h)
            val = -u * jv(n, u) / (v * kv(n, v))
            er_te = -(kv(n - 1, vr) - kv(n + 1, vr)) / 2 * fr * val
            er_tm = kvp(n, vr) * fr * val
            ep_te = kvp(n, vr) * fp * val
            ep_tm = -(kv(n - 1, vr) - kv(n + 1, vr)) / 2 * fp * val
            hr = -y_te * a * ep_te - y_tm * b * ep_tm
            hp = y_te * a * er_te + y_tm * b * er_tm
            hz = v / (1j * h * self.r) * y_te * a * kv(n, vr) * fp * val
        hx = hr * np.cos(p) - hp * np.sin(p)
        hy = hr * np.sin(p) + hp * np.cos(p)
        return np.array([hx, hy, hz])

    @staticmethod
    def u_jnu_jnpu_pec(num_n, num_m):
        us = np.empty((2, num_n, num_m))
        jnus = np.empty((2, num_n, num_m))
        jnpus = np.empty((2, num_n, num_m))
        for n in range(num_n):
            us[0, n] = jnp_zeros(n, num_m)
            us[1, n] = jn_zeros(n, num_m)
            jnus[0, n] = jv(n, us[0, n])
            jnus[1, n] = np.zeros(num_m)
            jnpus[0, n] = np.zeros(num_m)
            jnpus[1, n] = jvp(n, us[1, n])
        return us, jnus, jnpus

    def coefs_numpy(self, hs, w):
        As = []
        Bs = []
        for h, s, n, m in zip(hs, self.s_all, self.n_all, self.m_all):
            pol = "E" if s == 0 else "M"
            ai, bi = self.coef(h, w, (pol, n, m))
            As.append(ai)
            Bs.append(bi)
        return np.ascontiguousarray(As), np.ascontiguousarray(Bs)

    def coefs(self, hs, w):
        return coefs_cython(self, hs, w)

    def Ys(self, w, hs, As, Bs):
        vals = []
        for h, s, n, a, b in zip(hs, self.s_all, self.n_all, As, Bs):
            pol = "E" if s == 0 else "M"
            vals.append(self.Y(w, h, (pol, n, 1), a, b))
        return np.array(vals)

    # def Ymat(self, w, hs, As, Bs):
    #     mat = []
    #     for h1, s1, l1, n1, m1, a1, b1 in zip(
    #             hs, self.s_all, self.l_all, self.n_all, self.m_all, As, Bs):
    #         row = []
    #         for h2, s2, l2, n2, m2, a2, b2 in zip(
    #                 hs, self.s_all, self.l_all, self.n_all, self.m_all,
    #                 As, Bs):
    #             row.append(self.modes.Y_orthogonal(
    #                 w, h1, s1, l1, n1, m1, a1, b1,
    #                 h2, s2, l2, n2, m2, a2, b2))
    #         mat.append(row)
    #     return np.array(mat)

    def hAB(self, w):
        hs = np.array([self.beta(w, alpha) for alpha in self.alpha_all])
        As, Bs = self.coefs(hs, w)
        return hs, As, Bs

    def ABY(self, w, hs):
        e1 = self.fill(w)
        e2 = self.clad(w)
        return ABY_cython(
            w,
            self.r,
            self.s_all,
            self.n_all,
            self.m_all,
            hs,
            e1,
            e2,
            self.u_pec,
            self.jnu_pec,
            self.jnpu_pec,
        )

    # def ABYmat(self, w, hs):
    #     e1 = self.fill(w)
    #     e2 = self.clad(w)
    #     return ABYmat_cython(
    #         w, self.r, self.s_all, self.l_all,
    #         self.n_all, self.m_all, hs, e1, e2, self.u_pec, self.jnu_pec,
    #         self.jnpu_pec)

    def hABY(self, w):
        e1 = self.fill(w)
        e2 = self.clad(w)
        hs = np.array([self.beta(w, alpha) for alpha in self.alpha_all])
        As, Bs, Y = ABY_cython(
            w,
            self.r,
            self.s_all,
            self.n_all,
            self.m_all,
            hs,
            e1,
            e2,
            self.u_pec,
            self.jnu_pec,
            self.jnpu_pec,
        )
        return hs, As, Bs, Y

    def huvABY(self, w):
        e1 = self.fill(w)
        e2 = self.clad(w)
        hs = np.array([self.beta(w, alpha) for alpha in self.alpha_all])
        us, vs, As, Bs, Y = uvABY_cython(
            w,
            self.r,
            self.s_all,
            self.n_all,
            self.m_all,
            hs,
            e1,
            e2,
            self.u_pec,
            self.jnu_pec,
            self.jnpu_pec,
        )
        return hs, us, vs, As, Bs, Y

    # def hABYmat(self, w):
    #     e1 = self.fill(w)
    #     e2 = self.clad(w)
    #     hs = np.array([self.beta(w, alpha)
    #                    for alpha in self.alpha_all])
    #     As, Bs, Ymat = ABYmat_cython(
    #         w, self.r, self.s_all, self.l_all,
    #         self.n_all, self.m_all, hs, e1, e2, self.u_pec, self.jnu_pec,
    #         self.jnpu_pec)
    #     return hs, As, Bs, Ymat



from __future__ import annotations

import cmath
from typing import Dict, List, Tuple

import numpy as np
import ray
import riip
from scipy.optimize import root
from scipy.special import jn_zeros, jnp_zeros, jv, jvp, kv, kvp

from pymwm.utils.cylinder_utils import func_cython
from pymwm.waveguide import Sampling


class Samples(Sampling):
    """A class defining samples of phase constants of cylindrical waveguide
    modes.

    Attributes:
        r: A float indicating the radius of the circular cross section [um].
    """

    def __init__(self, size: float, fill: dict, clad: dict, params: dict):
        """Init Samples class.

        Args:
            size (float): The radius of the cross section [um]
            fill (dict): Parameters for riip.Material class for the core
            clad (dict): Parameters for riip.Material class for the clad
            params (dict): Keys and values are as follows:
                'wl_max' (float): The maximum wavelength [um].
                    Defaults to 5.0.
                'wl_min' (float): The minimum wavelength [um].
                    Defaults to 0.4.
                'wl_imag' (float): The minimum value of abs(c / f_imag) [um]
                    where f_imag is the imaginary part of the frequency.
                    Defaults to 5.0.
                'dw' (float): The frequency interval [rad c / 1um]=[2.99792458e14 rad / s].
                    Defaults to 1 / 64.
                'num_n' (int): The number of orders of modes.
                'num_m' (int): The number of modes in each order and polarization.
        """
        super().__init__(size, fill, clad, params)
        self.r = size

    @property
    def shape(self):
        return "cylinder"

    @property
    def num_all(self):
        return self.params["num_n"] * (2 * self.params["num_m"] + 1)

    def beta2_pec(self, w, n):
        """Return squares of phase constants for a PEC waveguide

        Args:
            w: A complex indicating the angular frequency
            n: A integer indicating the order of the modes.
        Returns:
            h2s: A 1D array indicating squares of phase constants, whose first
                num_m+1 elements are for TM-like modes and the rest are for
                TE-like modes.
        """
        w_comp = w.real + 1j * w.imag
        # The number of TM-like modes for each order is taken lager than
        # the number of TE-like modes since TM-like modes are almost identical
        # to those for PEC waveguide and can be calculated easily.
        # Dividing function by (x - x0) where x0 is already-found root
        # makes it easier to find new roots.
        num_m = self.params["num_m"]
        chi = jn_zeros(n, num_m + 1)
        h2s_mag = self.fill(w_comp) * w_comp ** 2 - chi ** 2 / self.r ** 2
        chi = jnp_zeros(n, num_m)
        h2s_elec = self.fill(w_comp) * w_comp ** 2 - chi ** 2 / self.r ** 2
        h2s = np.hstack((h2s_mag, h2s_elec))
        return h2s

    def u(self, h2: complex, w: complex, e1: complex) -> complex:
        return (1 + 1j) * cmath.sqrt(-0.5j * (e1 * w ** 2 - h2)) * self.r

    def v(self, h2: complex, w: complex, e2: complex) -> complex:
        return (1 - 1j) * cmath.sqrt(0.5j * (-e2 * w ** 2 + h2)) * self.r

    def eig_eq(
        self, h2: complex, w: complex, pol: str, n: int, e1: complex, e2: complex
    ):
        """Return the value of the characteristic equation

        Args:
            h2: The square of the phase constant.
            w: The angular frequency
            pol: The polarization
            n: The order of the modes
            e1: The permittivity of the core
            e2: The permittivity of the clad.
        Returns:
            val: A complex indicating the left-hand value of the characteristic
                equation.
        """
        h2comp = h2.real + 1j * h2.imag
        u = self.u(h2comp, w, e1)
        v = self.v(h2comp, w, e2)
        jus = jv(n, u)
        jpus = jvp(n, u)
        kvs = kv(n, v)
        kpvs = kvp(n, v)
        te = jpus / u + kpvs * jus / (v * kvs)
        tm = e1 * jpus / u + e2 * kpvs * jus / (v * kvs)
        if n == 0:
            if pol == "M":
                val = tm
            else:
                val = te
        else:
            val = (
                tm * te - h2comp * (n / w) ** 2 * ((1 / u ** 2 + 1 / v ** 2) * jus) ** 2
            )
        return val

    def beta2(
        self, w: complex, n: int, e1: complex, e2: complex, xis: np.ndarray
    ) -> tuple[np.ndarray, np.ndarray]:
        """Return roots and convergences of the characteristic equation

        Args:
            w (complex): Angular frequency.
            n (int): Order of the mode
            e1 (complex): Permittivity of tha core.
            e2 (complex): Permittivity of tha clad.
            xis (np.ndarray): Initial approximations for the roots
                whose number of elements is 2*num_m+1.
        Returns:
            xs: A 1D array indicating the roots, whose length is 2*num_m+1.
            success: A 1D array indicating the convergence information for xs.
        """
        if self.clad.label == "PEC":
            xs = self.beta2_pec(w, n)
            return xs, np.ones_like(xs, dtype=bool)
        num_m = self.params["num_m"]
        roots: list[complex] = []
        vals = []
        success: list[bool] = []

        for i, xi in enumerate(xis):
            if i < num_m + 1:
                pol = "M"
            else:
                pol = "E"
            result = root(
                func_cython,
                np.array([xi.real, xi.imag]),
                args=(w, pol, n, e1, e2, self.r, np.array(roots, dtype=complex)),
                method="hybr",
                options={"xtol": 1.0e-9},
            )
            x = result.x[0] + result.x[1] * 1j
            v = self.v(x, w, e2)
            if result.success:
                roots.append(x)
            if v.real > 0.0:
                success.append(result.success)
            else:
                success.append(False)
            vals.append(x)
        return np.array(vals), np.array(success)

    @staticmethod
    def beta_from_beta2(x):
        return (1 + 1j) * np.sqrt(-0.5j * x)

    def beta2_pec_per_mode(self, w, key):
        """Return squares of phase constants for a PEC waveguide

        Args:
            w: A complex indicating the angular frequency
            key: A tuple (pol, n, m) where
                pol: 'E' or 'M' indicating the polarization.
                n: A integer indicating the order of the mode.
                m: A integer indicating the ordinal of the mode in the same
                    order.
        Returns:
            h2s: A 1D array indicating squares of phase constants, whose first
                num_m+1 elements are for TM-like modes and the rest are for
                TE-like modes.
        """
        w_comp = w.real + 1j * w.imag
        pol, n, m = key
        if pol == "M":
            chi = jn_zeros(n, m)[-1]
            h2 = self.fill(w_comp) * w_comp ** 2 - chi ** 2 / self.r ** 2
        else:
            chi = jnp_zeros(n, m)[-1]
            h2 = self.fill(w_comp) * w_comp ** 2 - chi ** 2 / self.r ** 2
        return h2

    def jac(self, h2, args):
        """Return Jacobian of the characteristic equation

        Args:
            h2: A complex indicating the square of the phase constant.
            args: A tuple (w, n, e1, e2), where w indicates the angular
                frequency, n indicates  the order of the modes, e1 indicates
                the permittivity of the core, and e2 indicates the permittivity
                of the clad.
        Returns:
            val: A complex indicating the Jacobian of the characteristic
                equation.
        """
        w, pol, n, e1, e2 = args
        h2comp = h2.real + 1j * h2.imag
        u = self.u(h2comp, w, e1)
        v = self.v(h2comp, w, e2)
        jus = jv(n, u)
        jpus = jvp(n, u)
        kvs = kv(n, v)
        kpvs = kvp(n, v)
        du_dh2 = -self.r / (2 * u)
        dv_dh2 = self.r / (2 * v)
        te = jpus / u + kpvs * jus / (v * kvs)
        dte_du = -(
            u * (1 - n ** 2 / u ** 2) * jus + 2 * jpus
        ) / u ** 2 + jpus * kpvs / (v * kvs)
        dte_dv = (
            jus
            * (n ** 2 * kvs ** 2 / v + v * (kvs ** 2 - kpvs ** 2) - 2 * kvs * kpvs)
            / (v ** 2 * kvs ** 2)
        )
        tm = e1 * jpus / u + e2 * kpvs * jus / (v * kvs)
        dtm_du = e1 * dte_du
        dtm_dv = e2 * dte_dv
        if n == 0:
            if pol == "M":
                val = dtm_du * du_dh2 + dtm_dv * dv_dh2
            else:
                val = dte_du * du_dh2 + dte_dv * dv_dh2
        else:
            dre_dh2 = (
                -((n / w) ** 2)
                * jus
                * (
                    jus
                    * (
                        (1 / u ** 2 + 1 / v ** 2) ** 2
                        - self.r * h2comp * (1 / u ** 4 - 1 / v ** 4)
                    )
                    + jpus * 2 * h2comp * (1 / u ** 2 + 1 / v ** 2) ** 2
                )
            )
            val = (
                (dte_du * du_dh2 + dte_dv * dv_dh2) * tm
                + (dtm_du * du_dh2 + dtm_dv * dv_dh2) * te
                + dre_dh2
            )
        return val

    def beta2_w_min(self, n):
        """Return roots and convergences of the characteristic equation at
            the lowest angular frequency, ws[0].

        Args:
            n: A integer indicating the order of the mode
        Returns:
            xs: A 1D array indicating the roots, whose length is 2*num_m+1.
            success: A 1D array indicating the convergence information for xs.
        """
        if self.clad.label == "PEC":
            xs = self.beta2_pec(self.ws[0], n)
            success = np.ones_like(xs, dtype=bool)
            return xs, success
        w_0 = 0.1
        e1 = self.fill(w_0)
        e2_0 = -5.0e5 + self.clad(w_0).imag * 1j
        de2 = (self.clad(w_0) - e2_0) / 1000
        xis = xs = self.beta2_pec(w_0, n)
        success = np.ones_like(xs, dtype=bool)
        for i in range(1001):
            e2 = e2_0 + de2 * i
            xs, success = self.beta2(w_0, n, e1, e2, xis)
            for _, ok in enumerate(success):
                if not ok:
                    xs[_] = xis[_]
            xis = xs
        dw = (self.ws[0] - w_0) / 1000
        for i in range(1001):
            w = w_0 + dw * i
            e1 = self.fill(w)
            e2 = self.clad(w)
            xs, success = self.beta2(w, n, e1, e2, xis)
            for _, ok in enumerate(success):
                if not ok:
                    xs[_] = xis[_]
            xis = xs
        return xs, success

    def betas_convs(self, xs_success_list):
        num_n = self.params["num_n"]
        num_m = self.params["num_m"]
        betas = {}
        convs = {}
        for n in range(num_n):
            xs_array, success_array = xs_success_list[n]
            for m in range(1, num_m + 2):
                betas[("M", n, m)] = np.zeros(
                    (len(self.ws), len(self.wis)), dtype=complex
                )
                convs[("M", n, m)] = np.zeros((len(self.ws), len(self.wis)), dtype=bool)
            for m in range(1, num_m + 1):
                betas[("E", n, m)] = np.zeros(
                    (len(self.ws), len(self.wis)), dtype=complex
                )
                convs[("E", n, m)] = np.zeros((len(self.ws), len(self.wis)), dtype=bool)
            for iwi in range(len(self.wis)):
                for iwr in range(len(self.ws)):
                    w = self.ws[iwr] + 1j * self.wis[iwi]
                    e2 = self.clad(w)
                    for i in range(num_m + 1):
                        x = xs_array[iwr, iwi][i]
                        v = self.v(x, w, e2)
                        betas[("M", n, i + 1)][iwr, iwi] = self.beta_from_beta2(x)
                        convs[("M", n, i + 1)][iwr, iwi] = (
                            success_array[iwr, iwi][i]
                            if v.real > abs(v.imag)
                            else False
                        )
                    for i in range(num_m):
                        x = xs_array[iwr, iwi][i + num_m + 1]
                        v = self.v(x, w, e2)
                        betas[("E", n, i + 1)][iwr, iwi] = self.beta_from_beta2(x)
                        convs[("E", n, i + 1)][iwr, iwi] = (
                            success_array[iwr, iwi][i + num_m + 1]
                            if v.real > abs(v.imag)
                            else False
                        )
        return betas, convs

    def __call__(self, n: int):
        """Return a dict of the roots of the characteristic equation

        Args:
            n: A integer indicating the order of the mode
        Returns:
            betas: A dict containing arrays of roots, whose key is as follows:
                (pol, n, m):
                    pol: 'E' or 'M' indicating the polarization.
                    n: A integer indicating the order of the mode.
                    m: A integer indicating the ordinal of the mode in the same
                        order.
            convs: A dict containing the convergence information for betas,
                whose key is the same as above.
        """
        num_m = self.params["num_m"]
        xs_array = np.zeros((len(self.ws), len(self.wis), 2 * num_m + 1), dtype=complex)
        success_array = np.zeros(
            (len(self.ws), len(self.wis), 2 * num_m + 1), dtype=bool
        )
        iwr = iwi = 0
        wi = self.wis[iwi]
        xis, success = self.beta2_w_min(n)
        xs_array[iwr, iwi] = xis
        success_array[iwr, iwi] = success
        for iwr in range(1, len(self.ws)):
            wr = self.ws[iwr]
            w = wr + 1j * wi
            e1 = self.fill(w)
            e2 = self.clad(w)
            xs, success = self.beta2(w, n, e1, e2, xis)
            xs = np.where(success, xs, xis)
            xs_array[iwr, iwi] = xs
            success_array[iwr, iwi] = success
            xis = xs
        for iwi in range(1, len(self.wis)):
            wi = self.wis[iwi]
            for iwr in range(len(self.ws)):
                wr = self.ws[iwr]
                w = wr + 1j * wi
                e1 = self.fill(w)
                e2 = self.clad(w)
                if iwr == 0:
                    xis = xs_array[iwr, iwi - 1]
                else:
                    xis = (
                        xs_array[iwr, iwi - 1]
                        + xs_array[iwr - 1, iwi]
                        - xs_array[iwr - 1, iwi - 1]
                    )
                xs, success = self.beta2(w, n, e1, e2, xis)
                xs = np.where(success, xs, xis)
                xs_array[iwr, iwi] = xs
                success_array[iwr, iwi] = success
        return xs_array, success_array


class SamplesLowLoss(Samples):
    """A class defining samples of phase constants of cylindrical waveguide
    modes in a virtual low-loss clad waveguide by subclassing the CylinderSamples class.

    Attributes:
        fill: An instance of Material class for the core
        clad: An instance of Material class for the clad
        r: A float indicating the width of the slit [um].
        params: A dict whose keys and values are as follows:
            'wl_max': A float indicating the maximum wavelength [um]
            'wl_min': A float indicating the minimum wavelength [um]
            'wl_imag': A float indicating the minimum value of
                abs(c / f_imag) [um] where f_imag is the imaginary part of
                the frequency.
            'dw': A float indicating frequency interval
                [rad * c / 1um]=[2.99792458e14 rad / s].
            'num_n': An integer indicating the number of orders of modes.
    """

    def __init__(self, size: float, fill: dict, clad: dict, params: dict):
        """Init Samples class.

        size (float): The radius of the cross section [um]
        fill (dict): Parameters for riip.Material class for the core
        clad (dict): Parameters for riip.Material class for the clad
        params (dict): Keys and values are as follows:
            'wl_max' (float): The maximum wavelength [um].
                Defaults to 5.0.
            'wl_min' (float): The minimum wavelength [um].
                Defaults to 0.4.
            'wl_imag' (float): The minimum value of abs(c / f_imag) [um]
                where f_imag is the imaginary part of the frequency.
                Defaults to 5.0.
            'dw' (float): The frequency interval [rad c / 1um]=[2.99792458e14 rad / s].
                Defaults to 1 / 64.
            'num_n' (int): The number of orders of modes.
            'num_m' (int): The number of modes in each order and polarization.
        """
        super().__init__(size, fill, clad, params)

    def betas_convs(self, xs_success_list):
        num_iwr = len(self.ws)
        num_iwi = len(self.wis)
        num_n = self.params["num_n"]
        num_m = self.params["num_m"]
        betas = {}
        convs = {}
        for n in range(num_n):
            for m in range(1, num_m + 2):
                betas[("M", n, m)] = np.zeros(
                    (len(self.ws), len(self.wis)), dtype=complex
                )
                convs[("M", n, m)] = np.zeros((len(self.ws), len(self.wis)), dtype=bool)
            for m in range(1, num_m + 1):
                betas[("E", n, m)] = np.zeros(
                    (len(self.ws), len(self.wis)), dtype=complex
                )
                convs[("E", n, m)] = np.zeros((len(self.ws), len(self.wis)), dtype=bool)
        for iwr in range(num_iwr):
            for iwi in range(num_iwi):
                j = iwr * num_iwi + iwi
                w = self.ws[iwr] + 1j * self.wis[iwi]
                e2 = self.clad(w)
                for n in range(num_n):
                    for i in range(num_m + 1):
                        x = xs_success_list[j][0][n][i]
                        v = self.v(x, w, e2)
                        betas[("M", n, i + 1)][iwr, iwi] = self.beta_from_beta2(x)
                        convs[("M", n, i + 1)][iwr, iwi] = (
                            xs_success_list[j][1][n][i]
                            if v.real > abs(v.imag)
                            else False
                        )
                    for i in range(num_m):
                        x = xs_success_list[j][0][n][i + num_m + 1]
                        v = self.v(x, w, e2)
                        betas[("E", n, i + 1)][iwr, iwi] = self.beta_from_beta2(x)
                        convs[("E", n, i + 1)][iwr, iwi] = (
                            xs_success_list[j][1][n][i + num_m + 1]
                            if v.real > abs(v.imag)
                            else False
                        )
        return betas, convs


@ray.remote
class SamplesForRay(Samples):
    """A derived class in order to create ray actor."""

    def __init__(self, size: float, fill: dict, clad: dict, params: dict):
        super().__init__(size, fill, clad, params)

    def task(self, n: int):
        """Return a dict of the roots of the characteristic equation

        Args:
            n: A integer indicating the order of the mode
        Returns:
            betas: A dict containing arrays of roots, whose key is as follows:
                (pol, n, m):
                    pol: 'E' or 'M' indicating the polarization.
                    n: A integer indicating the order of the mode.
                    m: A integer indicating the ordinal of the mode in the same
                        order.
            convs: A dict containing the convergence information for betas,
                whose key is the same as above.
        """
        return super().__call__(n)


@ray.remote
class SamplesLowLossForRay(SamplesLowLoss):
    """A derived class in order to create ray actor."""

    def __init__(self, size: float, fill: dict, clad: dict, params: dict):
        super().__init__(size, fill, clad, params)

    def task(self, arg: Tuple[int, int, List[np.ndarray]]):
        """Return a dict of the roots of the characteristic equation

        Args:
            arg: (iwr, iwi, xis_list)
                iwr: The ordinal of the Re(w).
                iwi: The ordinal of the Im(w).
                xis_list: The initial guess of roots whose length is 2*num_m+1
        Returns:
            xs_list: A list of num_n 1D arrays indicating the roots, whose
                length is 2*num_m+1
            success_list: A list of num_n 1D arrays indicating the convergence
                information for xs, whose length is 2*num_m+1
        """
        iwr, iwi, xis_list = arg
        im_factor = self.clad.im_factor
        self.clad.im_factor = 1.0
        wr = self.ws[iwr]
        wi = self.wis[iwi]
        w = wr + 1j * wi
        e1 = self.fill(w)
        xs_list = []
        success_list = []
        for n, x0s in enumerate(xis_list):
            xis = xs = x0s
            success = np.ones_like(xs, dtype=bool)
            for i in range(1, 16):
                self.clad.im_factor = 0.7 ** i
                if i == 15 or self.clad.im_factor < im_factor:
                    self.clad.im_factor = im_factor
                e2 = self.clad(w)
                xs, success = self.beta2(w, n, e1, e2, xis)
                for _, ok in enumerate(success):
                    if not ok:
                        xs[_] = xis[_]
                xis = xs
            xs_list.append(xs)
            success_list.append(success)
        return xs_list, success_list



# -*- coding: utf-8 -*-

__author__ = "Munehiro Nishida"
__version__ = "0.2.3"
__license__ = "MIT"


def create(params):
    if params["core"]["shape"] == "cylinder":
        from .cylinder import Cylinder

        return Cylinder(params)
    elif params["core"]["shape"] == "slit":
        from .slit import Slit

        return Slit(params)
    else:
        raise ValueError(
            "Only 'cylinder' and 'slit' can be chosen for 'shape'" "at this time."
        )



from __future__ import annotations

import os
from typing import Dict, List, Tuple

import numpy as np
import pandas as pd
import scipy.optimize as so
import scipy.special as sp

from pymwm.utils.cutoff_utils import f_cython, f_fprime_cython, fprime_cython


class Cutoff:

    dirname = os.path.join(os.path.expanduser("~"), ".pymwm")
    filename = os.path.join(dirname, "cutoff.h5")

    def __init__(self, num_n: int, num_m: int) -> None:
        """Init Cutoff class."""
        self.num_n, self.num_m = num_n, num_m
        self.r_ratios = 0.001 * np.arange(1000)
        if os.path.exists(self.filename):
            samples = pd.read_hdf(self.filename)
            if self.num_n > samples["n"].max() + 1 or self.num_m > samples["m"].max():
                self.samples = self.cutoffs()
                self.samples.to_hdf(self.filename, "cutoff")
            else:
                self.samples = samples[
                    (samples["n"] < num_n) & (samples["m"] <= num_m)
                ].reset_index()
        if not os.path.exists(self.filename):
            if not os.path.exists(self.dirname):
                print("Folder Not Found.")
                os.mkdir(self.dirname)
            print("File Not Found.")
            self.samples = self.cutoffs()
            self.samples.to_hdf(self.filename, "cutoff")

    def __call__(self, alpha: tuple, r_ratio: float) -> float:
        df = self.samples
        df = df[(df["pol"] == alpha[0]) & (df["n"] == alpha[1]) & (df["m"] == alpha[2])]
        print(df)
        x = df["rr"].to_numpy()
        y = df["val"].to_numpy()
        print(x)
        i: int = np.where(x < r_ratio)[0][-1]
        a: float = r_ratio - x[i]
        b: float = x[i + 1] - r_ratio
        val: float = (b * y[i] + a * y[i + 1]) / (a + b)
        return val

    def PEC(
        self, x: float, r_ratio: float, n: int, pol: str
    ) -> tuple[float, float, float]:
        v = x * r_ratio

        def f_fp_fpp_fppp(func, z):
            f = func(n, z)
            fp = -func(n + 1, z) + n / z * f
            fpp = -fp / z - (1 - n ** 2 / z ** 2) * f
            fppp = -fpp / z - fp + (n ** 2 + 1) * fp / z ** 2 - 2 * n ** 2 / z ** 3 * f
            return f, fp, fpp, fppp

        jv, jpv, jppv, jpppv = f_fp_fpp_fppp(sp.jv, v)
        yv, ypv, yppv, ypppv = f_fp_fpp_fppp(sp.yv, v)
        jx, jpx, jppx, jpppx = f_fp_fpp_fppp(sp.jv, x)
        yx, ypx, yppx, ypppx = f_fp_fpp_fppp(sp.yv, x)
        if pol == "E":
            f = jpv * ypx - ypv * jpx
            fp = r_ratio * jppv * ypx + jpv * yppx - r_ratio * yppv * jpx - ypv * jppx
            fpp = (
                r_ratio ** 2 * jpppv * ypx
                + 2 * r_ratio * jppv * yppx
                + jpv * ypppx
                - r_ratio ** 2 * ypppv * jpx
                - 2 * r_ratio * yppv * jppx
                - ypv * jpppx
            )
        else:
            f = jv * yx - yv * jx
            fp = r_ratio * jpv * yx + jv * ypx - r_ratio * ypv * jx - yv * jpx
            fpp = (
                r_ratio ** 2 * jppv * yx
                + 2 * r_ratio * jpv * ypx
                + jv * yppx
                - r_ratio ** 2 * yppv * jx
                - 2 * ypv * jpx
                - yv * jppx
            )
        return f, fp, fpp

    def cutoffs_old(self):
        num_n, num_m = self.num_n, self.num_m
        z = []
        data = {}
        for pol in ["E", "M"]:
            for n in range(num_n):
                if pol == "E":
                    kini = sp.jnp_zeros(n, num_m)
                else:
                    kini = sp.jn_zeros(n, num_m)
                for m in range(1, num_m + 1):
                    data[(pol, n, m)] = kini[m - 1]
        z.append(data)
        for r_ratio in self.r_ratios[1:]:
            data = {}
            for pol in ["E", "M"]:
                for n in range(num_n):
                    for m in range(1, num_m + 1):
                        if r_ratio == self.r_ratios[1]:
                            ini_val = z[-1][(pol, n, m)]
                        else:
                            ini_val = z[-1][(pol, n, m)] * 2 - z[-2][(pol, n, m)]
                        sol = so.root_scalar(
                            self.PEC,
                            x0=ini_val,
                            fprime=True,
                            fprime2=True,
                            args=(r_ratio, n, pol),
                        )
                        kini = sol.root
                        data[(pol, n, m)] = kini
            z.append(data)
        df = pd.DataFrame()
        for i, (r_ratio, sample) in enumerate(zip(self.r_ratios, z)):
            for pol in ["E", "M"]:
                for n in range(self.num_n):
                    for m in range(1, self.num_m + 1):
                        df = df.append(
                            {
                                "irr": i,
                                "rr": r_ratio,
                                "pol": pol,
                                "n": n,
                                "m": m,
                                "val": sample[(pol, n, m)],
                            },
                            ignore_index=True,
                        )
        return df.astype({"irr": int, "n": int, "m": int})

    def cutoffs(self) -> pd.DataFrame:
        import ray

        if not ray.is_initialized():
            ray.init()
        rrs_id = ray.put(self.r_ratios)

        @ray.remote
        def func(alpha, kini, rrs):
            pol, n, m = alpha
            z = [kini]
            sol = so.root_scalar(
                f_fprime_cython,
                x0=kini,
                fprime=True,
                method="newton",
                args=(rrs[1], n, pol),
            )
            z = [kini, sol.root]
            for r_ratio in rrs[2:]:
                ini_val = z[-1] * 2 - z[-2]
                sol = so.root_scalar(
                    f_fprime_cython,
                    x0=ini_val,
                    fprime=True,
                    method="newton",
                    args=(r_ratio, n, pol),
                )
                z.append(sol.root)
            return z

        num_m = self.num_m
        args = []
        for n in range(self.num_n):
            for pol in ["M", "E"]:
                if pol == "E":
                    kinis = sp.jnp_zeros(n, num_m)
                else:
                    kinis = sp.jn_zeros(n, num_m)
                for m in range(1, num_m + 1):
                    kini = kinis[m - 1]
                    args.append(((pol, n, m), kini))

        result_ids = [func.remote(alpha, kini, rrs_id) for alpha, kini in args]
        results = ray.get(result_ids)
        if ray.is_initialized():
            ray.shutdown()
        df = pd.DataFrame()
        num_rr = len(self.r_ratios)
        for i in range(len(args)):
            (pol, n, m), _ = args[i]
            z = results[i]
            df1 = pd.DataFrame()
            df1["pol"] = np.full(num_rr, pol)
            df1["n"] = n
            df1["m"] = m
            df1["irr"] = np.arange(num_rr)
            df1["rr"] = self.r_ratios
            df1["val"] = z
            df = pd.concat([df, df1], ignore_index=True)
        return df



# -*- coding: utf-8 -*-
# cython: profile=False
cimport cython


@cython.cdivision(True)
def f_fprime_cython(double x, double r_ratio, int n, str pol):
    cdef:
        double v = x * r_ratio
        double jv, jpv, jppv, yv, ypv, yppv
        double jx, jpx, jppx, yx, ypx, yppx
        double vals[3]
        double f
        double fp

    if pol == "E":
        jn_jnp_jnpp(n, v, vals)
        jv = vals[0]
        jpv = vals[1]
        jppv = vals[2]

        yn_ynp_ynpp(n, v, vals)
        yv = vals[0]
        ypv = vals[1]
        yppv = vals[2]

        jn_jnp_jnpp(n, x, vals)
        jx = vals[0]
        jpx = vals[1]
        jppx = vals[2]

        yn_ynp_ynpp(n, x, vals)
        yx = vals[0]
        ypx = vals[1]
        yppx = vals[2]

        f = jpv * ypx - ypv * jpx
        fp = r_ratio * jppv * ypx + jpv * yppx - r_ratio * yppv * jpx - ypv * jppx
    else:
        jn_jnp(n, v, vals)
        jv = vals[0]
        jpv = vals[1]

        yn_ynp(n, v, vals)
        yv = vals[0]
        ypv = vals[1]

        jn_jnp(n, x, vals)
        jx = vals[0]
        jpx = vals[1]

        yn_ynp(n, x, vals)
        yx = vals[0]
        ypx = vals[1]

        f = jv * yx - yv * jx
        fp = r_ratio * jpv * yx + jv * ypx - r_ratio * ypv * jx - yv * jpx
    return f, fp


@cython.cdivision(True)
def f_cython(double x, double r_ratio, int n, str pol):
    cdef:
        double v = x * r_ratio

    if pol == "E":
        return jnp(n, v) * ynp(n, x) - ynp(n, v) * jnp(n, x)
    return  jn(n, v) * yn(n, x) - yn(n, v) * jn(n, x)


@cython.cdivision(True)
def fprime_cython(double x, double r_ratio, int n, str pol):
    cdef:
        double v = x * r_ratio
        double jv, jpv, jppv, yv, ypv, yppv
        double jx, jpx, jppx, yx, ypx, yppx
        double vals[3]

    jn_jnp_jnpp(n, v, vals)
    jv = vals[0]
    jpv = vals[1]
    jppv = vals[2]

    yn_ynp_ynpp(n, v, vals)
    yv = vals[0]
    ypv = vals[1]
    yppv = vals[2]

    jn_jnp_jnpp(n, x, vals)
    jx = vals[0]
    jpx = vals[1]
    jppx = vals[2]

    yn_ynp_ynpp(n, x, vals)
    yx = vals[0]
    ypx = vals[1]
    yppx = vals[2]

    if pol == "E":
        return r_ratio * jppv * ypx + jpv * yppx - r_ratio * yppv * jpx - ypv * jppx
    else:
        return r_ratio * jpv * yx + jv * ypx - r_ratio * ypv * jx - yv * jpx
